using System;
using System.Collections;
using System.Reflection;
using Kodama.Script.Engine;
using Kodama.Script.Factory;
using Kodama.Script.Engine.CodeDom.CSharpScript;
using Kodama.Script.Engine.CodeDom.JScript;
using Kodama.Script.Engine.CodeDom.VBScript;

using Kodama.DependencyInjection.Container;

namespace Kodama.DependencyInjection.Builder
{
    /// <summary>
    /// XNvg𗘗pāADependencyInjectionRei𐶐
    /// R|[lgo^ÃCX^XԂr_
    /// </summary>
    /// <seealso cref="Kodama.DependencyInjection.Container.ComponentContainerImpl"/>
    public class ScriptComponentContainerBuilder : IComponentContainerBuilder
    {
        /// <summary>
        /// XNvgt@C̃pX
        /// </summary>
        private string scriptPath = null;

        /// <summary>
        /// XNvgQƂAZu̖O
        /// </summary>
        private ArrayList assemblyNames = new ArrayList();

        /// <summary>
        /// XNvgt@C̃pX
        /// </summary>
        public string Script
        {
            set { scriptPath = value; }
        }

        /// <summary>
        /// XNvgQƂAZu̒ǉ
        /// </summary>
        /// <param name="assemblyName">XNvgQƂAZu</param>
        public void AddReference(string assemblyName)
        {
            assemblyNames.Add(assemblyName);
        }

        /// <summary>
        /// XNvg𗘗pāADependencyInjectionRei𐶐
        /// R|[lgo^ÃCX^XԂ
        /// </summary>
        /// <returns>ꂽDependencyInjectionRei</returns>
        /// <exception cref="CompileErrorException">
        /// XNvg̃RpCG[ꍇ̗͂O𓊂
        /// </exception>
        /// <exception cref="FactoryNotFoundException">
        /// DependencyInjectionReĩt@Ng[\bhXNvg
        /// ݂Ȃꍇ̗͂O𓊂
        /// </exception>
        /// <remarks>
        /// <p>XNvgɂC#AVBScriptJScriptpł܂BXNvg
        /// ComponentContainerFactoryAttributeIComponentContainerC^[tFCX
        /// ƌ݊̂lԂStaticȊ֐ADependencyInjectionRei𐶐
        /// ˑ֌Wݒ肵̂Ԃt@Ng[\bhƂȂ܂BXNvg
        /// ̃t@Ng[\bh݂ȂꍇFactoryNotFoundExceptionO
        /// ܂BȉɃXNvg̗܂B</p>
        /// <code lang="C#">
        /// using System;
        /// using Kodama.DependencyInjection.Container;
        /// using Kodama.DependencyInjection.Factory;
        /// using HogeHoge;
        ///
        /// public class ComponentContainerBuilder
        /// {
        ///     [ComponentContainerFactory]
        ///     public IMutableComponentContainer Build()
        ///     {
        ///         IMutableComponentContainer container = new ComponentContainerImpl();
        ///         container.Register(typeof(ClassA));
        ///         container.Register(typeof(ClassB));
        ///         return container;
        ///     }
        /// }
        /// </code>
        /// </remarks>
        public IComponentContainer Build()
        {
            IScriptEngineFactory[] factories = new IScriptEngineFactory[]
                {new CSharpScriptEngineFactory(), new JScriptEngineFactory(), new VBScriptEngineFactory()};
            foreach (IScriptEngineFactory factory in factories)
            {
                if (!factory.CanCompile(scriptPath))
                {
                    continue;
                }

                IScriptEngine engine = factory.CreateFromFile(scriptPath);

                engine.AddReference(Assembly.GetExecutingAssembly().Location);
                foreach (string assemblyName in assemblyNames)
                {
                    engine.AddReference(assemblyName);
                }

                try 
                {
                    engine.Compile();
                }
                finally
                {
                    engine.Close();
                }

                foreach (Type type in engine.ScriptAssembly.GetTypes())
                {
                    foreach (MethodInfo mi in type.GetMethods())
                    {
                        if (!mi.IsStatic)
                        {
                            continue;
                        }
                        if (mi.GetParameters().Length != 0)
                        {
                            continue;
                        }
                        if (!typeof(IComponentContainer).IsAssignableFrom(mi.ReturnType))
                        {
                            continue;
                        }
                        if (!Attribute.IsDefined(mi, typeof(ComponentContainerFactoryAttribute)))
                        {
                            continue;
                        }

                        IComponentContainer container = (IComponentContainer)mi.Invoke(null, null);

                        engine.Close();

                        return container;
                    }
                }

                engine.Close();

                throw new FactoryNotFoundException();
            }

            throw new NotSupportedScriptException();
        }
    }
}
