using System;
using System.Collections;
using Kodama.DependencyInjection.Component;

namespace Kodama.DependencyInjection.Container
{
    /// <summary>
    /// DependencyInjectionRei̎NXł
    /// </summary>
    /// <example>
    /// <code lang="C#">
    /// using System;
    /// using Kodama.DependencyInjection.Component;
    /// using Kodama.DependencyInjection.Container;
    /// using Kodama.DependencyInjection.Factory;
    /// using Kodama.Function;
    /// using Kodama.Function.Bind;
    /// using Kodama.Function.Member;
    ///
    /// ....
    ///
    /// public interface IBar
    /// {
    ///     void Print();
    /// }
    ///
    /// public class BarImpl : IBar
    /// {
    ///     public void Print()
    ///     {
    ///         Console.WriteLine("BarImpl");
    ///     }
    /// }
    ///
    /// public class Foo
    /// {
    ///     private IBar dependency1;
    ///     private IBar dependency2;
    ///     private int  val;
    ///
    ///     // InjectionPoint̂\bh͎I
    ///     // Zb^[CWFNṼ\bhɐݒ肳
    ///     [InjectionPoint]
    ///     public void SetDependency1(IBar dep)
    ///     {
    ///         dependency1 = dep;
    ///     }
    ///
    ///     public void SetDependency2(IBar dep, int v)
    ///     {
    ///         dependency2 = dep;
    ///         val         = v;
    ///     }
    ///
    ///     // InitializationPoint̂\bh͎I
    ///     // \bhɐݒ肳
    ///     [InitializationPoint]
    ///     public void Initialize1()
    ///     {
    ///         Console.WriteLine("Init1");
    ///     }
    ///
    ///     public void Initialize2(int v)
    ///     {
    ///         Console.WriteLine("Init2 val = " + v.ToString());
    ///     }
    /// }
    ///
    /// ...
    ///
    /// // ʏ̓Reiւ̃R|[lg̓o^AZb^[CWFNVy
    /// // \bh̐ݒ̓XNvgōsB
    /// // ڍׂKodama.DependencyInjection.Factory.DefaultComponetContainerFactory#Create
    /// // \bh̃I[o[[hQƁB
    ///
    /// IMutableComponentContainer container = new ComponentContainerImpl();
    ///
    /// continer.Register(typeof(BarImpl));
    ///
    /// IComponentEntry entry = new PrototypeComponetEntry(container, typeof(Foo));
    /// 
    /// // 蓮ɂZb^[CWFNV̐ݒ
    /// entry.AddInjectionFanctor(
    ///     new BindFunctor(
    ///         new MemberFunctor(typeof(Foo).GetMethod("SetDependency2")),
    ///         new NotBoundArgument(0),
    ///         new TypedArgumentComponentProvider(container, typeof(IBar)),
    ///         1));
    ///
    /// // 蓮ɂ鏉\bh̐ݒ
    /// entry.AddInitializationFactor(
    ///     new BindFunctor(
    ///         new MemberFunctor(typeof(Foo).GetMethod("Initialize2")),
    ///         new NotBoundArgument(0),
    ///         2));
    ///
    /// container.Register(entry);
    ///
    /// Foo foo = (Foo)continer.GetComponent(typeof(Foo));
    /// </code>
    /// </example>
    public class ComponentContainerImpl : IMutableComponentContainer
    {
        /// <summary>
        /// ̃C^[tFCXɗD悵Ċ蓖ĂR|[lg̏
        /// </summary>
        private Hashtable primaryComponentEntries = new Hashtable();

        /// <summary>
        /// o^ĂR|[lg̏
        /// </summary>
        private ArrayList secondaryComponentEntries = new ArrayList();

        /// <summary>
        /// qRei
        /// </summary>
        private ArrayList children = new ArrayList();

        /// <summary>
        /// R|[lg擾
        /// </summary>
        /// <param name="componentType">擾R|[lǧ^</param>
        /// <returns>ɂ킽ꂽ^̃R|[lg</returns>
        /// <exception cref="CyclicDependencyException">
        /// ˑ֌WzĂƂɔO
        /// </exception>
        /// <exception cref="TooManyRegistrationException">
        /// w肳ꂽC^[tFCXR|[lgꍇɔO
        /// </exception>
        /// <exception cref="ComponentNotFoundException">
        /// w肳ꂽC^[tFCXR|[lgȂꍇɔO
        /// </exception>
        /// <exception cref="CyclicDependencyException">
        /// R|[lg̈ˑ֌WzĂĂAzĂR|[lg
        /// Prototype[h̏ꍇɔO
        /// </exception>
        public object GetComponent(Type componentType)
        {
            return GetComponentEntry(componentType).GetInstance();
        }

        /// <summary>
        /// R|[lg擾
        /// </summary>
        /// <param name="componentName">擾R|[lg̖O</param>
        /// <returns>ɂ킽ꂽÕR|[lg</returns>
        /// <exception cref="TooManyRegistrationException">
        /// w肳ꂽÕR|[lgꍇɔO
        /// </exception>
        /// <exception cref="ComponentNotFoundException">
        /// w肳ꂽÕR|[lgȂꍇɔO
        /// </exception>
        /// <exception cref="CyclicDependencyException">
        /// R|[lg̈ˑ֌WzĂĂAzĂR|[lg
        /// Prototype[h̏ꍇɔO
        /// </exception>
        public object GetComponent(string componentName)
        {
            return GetComponentEntry(componentName).GetInstance();
        }

        /// <summary>
        /// R|[lgo^Ă邩`FbN
        /// </summary>
        /// <param name="componentType">o^Ă邩mFR|[lǧ^</param>
        /// <returns>R|[lgo^Ă邩ǂ</returns>
        public bool Contains(Type componentType)
        {
            foreach (IComponentEntry entry in primaryComponentEntries.Values)
            {
                if (object.Equals(componentType, entry.ComponentType))
                {
                    return true;
                }
            }
            foreach (IComponentEntry entry in secondaryComponentEntries)
            {
                if (object.Equals(componentType, entry.ComponentType))
                {
                    return true;
                }
            }
            foreach (IComponentContainer child in children)
            {
                if (child.Contains(componentType))
                {
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// R|[lgo^Ă邩`FbN
        /// </summary>
        /// <param name="componentName">o^Ă邩mFR|[lg̖O</param>
        /// <returns>R|[lgo^Ă邩ǂ</returns>
        public bool Contains(string componentName)
        {
            foreach (IComponentEntry entry in primaryComponentEntries.Values)
            {
                if (string.Compare(componentName, entry.ComponentName) == 0) 
                {
                    return true;
                }
            }
            foreach (IComponentEntry entry in secondaryComponentEntries)
            {
                if (string.Compare(componentName, entry.ComponentName) == 0) 
                {
                    return true;
                }
            }
            foreach (IComponentContainer child in children)
            {
                if (child.Contains(componentName))
                {
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// R|[lgo^
        /// </summary>
        /// <param name="componentType">o^R|[lǧ^</param>
        public void Register(Type componentType)
        {
            Register(new PrototypeComponentEntry(this, componentType));
        }

        /// <summary>
        /// R|[lgo^
        /// </summary>
        /// <param name="componentInstance">o^R|[lg̃CX^X</param>
        public void RegisterInstance(object componentInstance)
        {
            Register(new OuterComponentEntry(this, componentInstance));
        }

        /// <summary>
        /// R|[lgo^
        /// </summary>
        /// <param name="componentType">o^R|[lǧ^</param>
        /// <param name="componentName">o^R|[lg̖O</param>
        public void Register(Type componentType, string componentName)
        {
            Register(new PrototypeComponentEntry(this, componentType, componentName));
        }

        /// <summary>
        /// R|[lgo^
        /// </summary>
        /// <param name="componentInstance">o^R|[lg̃CX^X</param>
        /// <param name="componentName">o^R|[lg̖O</param>
        public void RegisterInstance(object componentInstance, string componentName)
        {
            Register(new OuterComponentEntry(this, componentInstance, componentName));
        }

        /// <summary>
        /// R|[lgo^
        /// </summary>
        /// <param name="componentEntry">o^R|[lg̏</param>
        public void Register(IComponentEntry componentEntry)
        {
            secondaryComponentEntries.Add(componentEntry);
        }

        /// <summary>
        /// R|[lgo^
        /// </summary>
        /// <remarks>
        /// ̃C^[tFCXɗD悵Ċ蓖ĂR|[lgw肵܂B
        /// </remarks>
        /// <param name="interfaceType">D悵Ċ蓖ĂC^[tFCX</param>
        /// <param name="implementComponentType">D悵Ċ蓖ĂR|[lg</param>
        /// <exception cref="TooManyRegistrationException">
        /// w肳ꂽC^[tFCXR|[lgꍇɔO
        /// </exception>
        public void Register(Type interfaceType, Type implementComponentType)
        {
            Register(interfaceType, new PrototypeComponentEntry(this, implementComponentType));
        }

        /// <summary>
        /// R|[lgo^
        /// </summary>
        /// <remarks>
        /// ̃C^[tFCXɗD悵Ċ蓖ĂR|[lgw肵܂B
        /// </remarks>
        /// <param name="interfaceType">D悵Ċ蓖ĂC^[tFCX</param>
        /// <param name="implementComponentInstance">D悵Ċ蓖ĂR|[lg̃CX^X</param>
        public void RegisterInstance(Type interfaceType, object implementComponentInstance)
        {
            Register(interfaceType, new OuterComponentEntry(this, implementComponentInstance));
        }

        /// <summary>
        /// R|[lgo^
        /// </summary>
        /// <remarks>
        /// ̃C^[tFCXɗD悵Ċ蓖ĂR|[lgw肵܂B
        /// </remarks>
        /// <param name="interfaceType">D悵Ċ蓖ĂC^[tFCX</param>
        /// <param name="implementComponentType">D悵Ċ蓖ĂR|[lǧ^</param>
        /// <param name="componentName">D悵Ċ蓖ĂR|[lg̖O</param>
        /// <exception cref="TooManyRegistrationException">
        /// w肳ꂽC^[tFCXR|[lgꍇɔO
        /// </exception>
        public void Register(Type interfaceType, Type implementComponentType, string componentName)
        {
            Register(interfaceType, new PrototypeComponentEntry(this, implementComponentType, componentName));
        }

        /// <summary>
        /// R|[lgo^
        /// </summary>
        /// <remarks>
        /// ̃C^[tFCXɗD悵Ċ蓖ĂR|[lgw肵܂B
        /// </remarks>
        /// <param name="interfaceType">D悵Ċ蓖ĂC^[tFCX</param>
        /// <param name="implementComponentInstance">D悵Ċ蓖ĂR|[lg̃CX^X</param>
        /// <param name="componentName">D悵Ċ蓖ĂR|[lg̖O</param>
        public void RegisterInstance(Type interfaceType, object implementComponentInstance, string componentName)
        {
            Register(interfaceType, new OuterComponentEntry(this, implementComponentInstance, componentName));
        }

        /// <summary>
        /// R|[lgo^
        /// </summary>
        /// <remarks>
        /// ̃C^[tFCXɗD悵Ċ蓖ĂR|[lgw肵܂B
        /// </remarks>
        /// <param name="interfaceType">D悵Ċ蓖ĂC^[tFCX</param>
        /// <param name="implementComponentEntry">D悵Ċ蓖ĂR|[lg̏</param>
        /// <exception cref="TooManyRegistrationException">
        /// w肳ꂽC^[tFCXR|[lgꍇɔO
        /// </exception>
        public void Register(Type interfaceType, IComponentEntry implementComponentEntry)
        {
            if (primaryComponentEntries.Contains(interfaceType))
            {
                throw new TooManyRegistrationException();
            }

            primaryComponentEntries.Add(interfaceType, implementComponentEntry);
        }

        /// <summary>
        /// w肳ꂽC^[tFCXR|[lg̓o^擾
        /// </summary>
        /// <param name="interfaceType">擾R|[lg̃C^[tFCX</param>
        /// <returns>w肳ꂽC^[tFCXR|[lg̓o^</returns>
        /// <exception cref="TooManyRegistrationException">
        /// w肳ꂽC^[tFCXR|[lgꍇɔO
        /// </exception>
        /// <exception cref="ComponentNotFoundException">
        /// w肳ꂽC^[tFCXR|[lgȂꍇɔO
        /// </exception>
        public IComponentEntry GetComponentEntry(Type interfaceType)
        {
            if (primaryComponentEntries.Contains(interfaceType))
            {
                return (IComponentEntry)primaryComponentEntries[interfaceType];
            }

            if (CheckTooManyRegistration(interfaceType))
            {
                throw new TooManyRegistrationException();
            }

            foreach (IComponentEntry entry in secondaryComponentEntries)
            {
                if (interfaceType.IsAssignableFrom(entry.ComponentType))
                {
                    return entry;
                }
            }

            foreach (IComponentContainer child in children)
            {
                try
                {
                    return child.GetComponentEntry(interfaceType);
                }
                catch (ComponentNotFoundException)
                {
                    // Ȃ
                }
            }

            throw new ComponentNotFoundException();
        }

        /// <summary>
        /// w肳ꂽOR|[lg̓o^擾
        /// </summary>
        /// <param name="componentName">擾R|[lg̖O</param>
        /// <returns>w肳ꂽOR|[lg̓o^</returns>
        /// <exception cref="TooManyRegistrationException">
        /// w肳ꂽOR|[lgꍇɔO
        /// </exception>
        /// <exception cref="ComponentNotFoundException">
        /// w肳ꂽOR|[lgȂꍇɔO
        /// </exception>
        public IComponentEntry GetComponentEntry(string componentName)
        {
            if (CheckTooManyRegistration(componentName))
            {
                throw new TooManyRegistrationException();
            }

            foreach (IComponentEntry entry in primaryComponentEntries.Values)
            {
                if (string.Compare(componentName, entry.ComponentName) == 0)
                {
                    return entry;
                }
            }

            foreach (IComponentEntry entry in secondaryComponentEntries)
            {
                if (string.Compare(componentName, entry.ComponentName) == 0)
                {
                    return entry;
                }
            }

            foreach (IComponentContainer child in children)
            {
                try
                {
                    return child.GetComponentEntry(componentName);
                }
                catch (ComponentNotFoundException)
                {
                    // Ȃ
                }
            }

            throw new ComponentNotFoundException();
        }

        /// <summary>
        /// w肳ꂽC^[tFCXR|[lgo^ĂȂ`FbN
        /// </summary>
        /// <param name="interfaceType">`FbNC^[tFCX</param>
        /// <returns>w肳ꂽC^[tFCXR|[lgłɕo^ĂȂǂ</returns>
        private bool CheckTooManyRegistration(Type interfaceType)
        {
            int count = 0;
            foreach (IComponentEntry entry in secondaryComponentEntries)
            {
                if (interfaceType.IsAssignableFrom(entry.ComponentType)) 
                {
                    ++count;
                }
            }

            // gbL[ȃJEg̎d...
            // qIComponentContainer#GetComponentEntryĂяoƂ
            // qIComponentContainerImpl#CheckTooManyRegistration
            // deXgAdΗO𓊂
            foreach (IComponentContainer child in children)
            {
                try
                {
                    child.GetComponentEntry(interfaceType);
                    ++count;
                }
                catch (ComponentNotFoundException)
                {
                    // Ȃ
                }
                catch (TooManyRegistrationException)
                {
                    return true;
                }
            }

            return count > 1;
        }

        /// <summary>
        /// OR|[lgo^ĂȂ`FbN
        /// </summary>
        /// <param name="componentName">R|[lg̖O</param>
        /// <returns>OR|[lgłɕo^ĂȂǂ</returns>
        private bool CheckTooManyRegistration(string componentName)
        {
            int count = 0;
            foreach (IComponentEntry entry in primaryComponentEntries.Values)
            {
                if (string.Compare(componentName, entry.ComponentName) == 0) 
                {
                    ++count;
                }
            }
            foreach (IComponentEntry entry in secondaryComponentEntries)
            {
                if (string.Compare(componentName, entry.ComponentName) == 0) 
                {
                    ++count;
                }
            }

            // gbL[ȃJEg̎d...
            // qIComponentContainer#GetComponentEntryĂяoƂ
            // qIComponentContainerImpl#CheckTooManyRegistration
            // deXgAdΗO𓊂
            foreach (IComponentContainer child in children)
            {
                try
                {
                    child.GetComponentEntry(componentName);
                    ++count;
                }
                catch (ComponentNotFoundException)
                {
                    // Ȃ
                }
                catch (TooManyRegistrationException)
                {
                    return true;
                }
            }

            return count > 1;
        }

        /// <summary>
        /// o^ĂR|[lg̃CX^Xj
        /// </summary>
        /// <remarks>
        /// o^ĂR|[lĝSingletonŃCX^X
        /// ǗĂ̂̃CX^XjBjR|[lg
        /// IDisposableĂ邱ƁB
        /// </remarks>
        public void Discard()
        {
            foreach (IComponentEntry entry in primaryComponentEntries.Values)
            {
                entry.Discard();
            }
            foreach (IComponentEntry entry in secondaryComponentEntries)
            {
                entry.Discard();
            }
        }

        /// <summary>
        /// ׂĂ̎qRei擾
        /// </summary>
        /// <returns>qRei̔z</returns>
        public IComponentContainer[] GetChildren()
        {
            return (IComponentContainer[])children.ToArray(typeof(IComponentContainer));
        }

        /// <summary>
        /// qReiǉ
        /// </summary>
        /// <param name="child">ǉqRei</param>
        public void AddChild(IComponentContainer child)
        {
            children.Add(child);
        }
    }
}
